/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "DocumentControl.h"

namespace
{
  struct SignalInfo {
    DocumentControl* doc_ctrl;
    C4sDocumentWindow* doc_window;
  };
  
  void destroy_info(gpointer data, GObject* where_the_object_was)
  {
    delete static_cast<SignalInfo*>(data);
  }
}

DocumentControl::DocumentControl(MainMenu& menu, FileHolder& file_holder, C4sPreferences* preferences, Gtk::Window& parent):
  m_menu(menu), m_file_holder(file_holder), m_preferences(preferences), m_parent(parent), m_preferences_dialog(NULL)
{
  file_holder.signal_document_added().connect(sigc::mem_fun(*this, &DocumentControl::on_document_added));
  file_holder.signal_document_changed().connect(sigc::mem_fun(*this, &DocumentControl::on_document_changed));
  menu.action_edit_undo->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_undo));
  menu.action_edit_redo->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_redo));

  menu.action_edit_cut->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_cut));
  menu.action_edit_copy->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_copy));
  menu.action_edit_paste->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_paste));

  menu.action_edit_find->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_find));
  menu.action_edit_replace->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_replace));
  menu.action_edit_goto->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_goto));

  menu.action_edit_preferences->signal_activate().connect(sigc::mem_fun(*this, &DocumentControl::on_preferences));
}

DocumentControl::~DocumentControl()
{
  if(m_preferences_dialog != NULL)
    gtk_widget_destroy(GTK_WIDGET(m_preferences_dialog));
}

void DocumentControl::on_document_added(C4sDocumentWindow* window)
{
  SignalInfo* info = new SignalInfo;
  info->doc_ctrl = this;
  info->doc_window = window;
  
  // Frees info when the source buffer is destroyed
  g_object_weak_ref(G_OBJECT(c4s_document_window_get_document(window)), &destroy_info, info);

  g_signal_connect(G_OBJECT(c4s_document_window_get_document(window)), "notify::can-undo", (GCallback)&DocumentControl::on_can_undo_static, info);
  g_signal_connect(G_OBJECT(c4s_document_window_get_document(window)), "notify::can-redo", (GCallback)&DocumentControl::on_can_redo_static, info);
}

void DocumentControl::on_document_changed(C4sDocumentWindow* window)
{
  if(window == NULL)
  {
    m_menu.action_edit_undo->set_sensitive(false);
    m_menu.action_edit_redo->set_sensitive(false);
  }
  else
  {
    m_menu.action_edit_undo->set_sensitive(gtk_source_buffer_can_undo(GTK_SOURCE_BUFFER(c4s_document_window_get_document(window))) == TRUE);
    m_menu.action_edit_redo->set_sensitive(gtk_source_buffer_can_redo(GTK_SOURCE_BUFFER(c4s_document_window_get_document(window))) == TRUE);
  }

  m_menu.action_edit_cut->set_sensitive(window != NULL);
  m_menu.action_edit_copy->set_sensitive(window != NULL);
  m_menu.action_edit_paste->set_sensitive(window != NULL);
}

void DocumentControl::on_undo()
{
  gtk_source_buffer_undo(GTK_SOURCE_BUFFER(c4s_document_window_get_document(m_file_holder.get_current_document())));
}

void DocumentControl::on_redo()
{
  gtk_source_buffer_redo(GTK_SOURCE_BUFFER(c4s_document_window_get_document(m_file_holder.get_current_document())));
}

void DocumentControl::on_cut()
{
  gtk_text_buffer_cut_clipboard(GTK_TEXT_BUFFER(c4s_document_window_get_document(m_file_holder.get_current_document())), gtk_clipboard_get(GDK_SELECTION_CLIPBOARD), TRUE);
}

void DocumentControl::on_copy()
{
  gtk_text_buffer_copy_clipboard(GTK_TEXT_BUFFER(c4s_document_window_get_document(m_file_holder.get_current_document())), gtk_clipboard_get(GDK_SELECTION_CLIPBOARD));
}

void DocumentControl::on_paste()
{
  gtk_text_buffer_paste_clipboard(GTK_TEXT_BUFFER(c4s_document_window_get_document(m_file_holder.get_current_document())), gtk_clipboard_get(GDK_SELECTION_CLIPBOARD), NULL, TRUE);
}

void DocumentControl::on_find()
{
  if(m_find_dialog.get() == NULL)
    m_find_dialog.reset(new FindDialog(m_parent, m_file_holder));

  m_find_dialog->set_search_only(true);
  m_find_dialog->present();
}

void DocumentControl::on_replace()
{
  if(m_find_dialog.get() == NULL)
    m_find_dialog.reset(new FindDialog(m_parent, m_file_holder));

  m_find_dialog->set_search_only(false);
  m_find_dialog->present();
}

void DocumentControl::on_goto()
{
  if(m_goto_dialog.get() == NULL)
    m_goto_dialog.reset(new GotoDialog(m_parent, m_file_holder));

  m_goto_dialog->present();
}

void DocumentControl::on_preferences_destroy()
{
  m_preferences_dialog = NULL;
}

void DocumentControl::on_preferences()
{
  if(m_preferences_dialog == NULL)
  {
    m_preferences_dialog = c4s_preferences_dialog_new(GTK_WINDOW(m_parent.gobj()), GTK_DIALOG_NO_SEPARATOR, m_preferences, m_file_holder.get_engine_function_list());

    g_signal_connect(
      G_OBJECT(m_preferences_dialog),
      "destroy",
      G_CALLBACK(on_preferences_destroy_static),
      this
    );
  }

  gtk_window_present(GTK_WINDOW(m_preferences_dialog));
}

void DocumentControl::on_can_undo_static(GObject* object, GParamSpec* pspec, gpointer info)
{
  SignalInfo* sig_info = static_cast<SignalInfo*>(info);

  if(sig_info->doc_ctrl->m_file_holder.get_current_document() == sig_info->doc_window)
    sig_info->doc_ctrl->m_menu.action_edit_undo->set_sensitive(gtk_source_buffer_can_undo(GTK_SOURCE_BUFFER(object)) == TRUE);
}

void DocumentControl::on_can_redo_static(GObject* object, GParamSpec* pspec, gpointer info)
{
  SignalInfo* sig_info = static_cast<SignalInfo*>(info);

  if(sig_info->doc_ctrl->m_file_holder.get_current_document() == sig_info->doc_window)
    sig_info->doc_ctrl->m_menu.action_edit_redo->set_sensitive(gtk_source_buffer_can_redo(GTK_SOURCE_BUFFER(object)) == TRUE);
}
