/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-i18n.h>

#include <stdexcept>
#include <fstream>

#ifdef _WIN32
# include <windows.h>
#endif

#include "FileHolder.h"
#include <gtkmm/box.h>
#include <gtkmm/button.h>
#include <gtkmm/image.h>
#include <gtkmm/stock.h>
#include <gtkmm/window.h>
#include <gtkmm/messagedialog.h>

namespace
{
  class TabLabel: public Gtk::HBox
  {
  public:
    typedef Glib::SignalProxy0<void> SignalClicked;
    TabLabel(C4sDocument* document);

    SignalClicked signal_close() { return m_button.signal_clicked(); }
  private:
    static void on_modified_changed_static(C4sDocument* document, gpointer user_data) { static_cast<TabLabel*>(user_data)->on_modified_changed(); }
    static void on_title_changed_static(C4sDocument* document, const gchar* title, gpointer user_data) { static_cast<TabLabel*>(user_data)->on_title_changed(); }

    void on_title_changed();
    void on_modified_changed();

    void update_label();

    static void tab_label_style_set_cb(GtkWidget* self,
                                       GtkStyle* previous_style,
				       gpointer user_data)
    {
      gint w, h;
      gtk_icon_size_lookup_for_settings(gtk_widget_get_settings(self), GTK_ICON_SIZE_MENU, &w, &h);
      GtkWidget* button = GTK_WIDGET(static_cast<TabLabel*>(user_data)->m_button.gobj());
      gtk_widget_set_size_request(button, w + 2, h + 2);
    }

    C4sDocument* m_document;

    Gtk::Label m_label;
    Gtk::Button m_button;
    Gtk::Image m_image;
  };

  TabLabel::TabLabel(C4sDocument* document):
    Gtk::HBox(false, 4),
    m_document(document),
    m_image(Gtk::Stock::CLOSE, Gtk::ICON_SIZE_MENU)
  {
    // Set correct label
    update_label();

    // Lookup icon size
    int width, height;
    Gtk::IconSize::lookup(Gtk::ICON_SIZE_MENU, width, height);

    m_button.set_border_width(0);
    set_border_width(0);

    /* Code stolen from gedit */
    /* don't allow focus on the close button */
    gtk_button_set_focus_on_click (GTK_BUTTON (m_button.gobj()), FALSE);

    /* make it as small as possible */
    GtkRcStyle* rcstyle = gtk_rc_style_new ();
    rcstyle->xthickness = rcstyle->ythickness = 0;
    gtk_widget_modify_style (GTK_WIDGET(m_button.gobj()), rcstyle);
    gtk_rc_style_unref (rcstyle);

    gtk_misc_set_padding(GTK_MISC(m_label.gobj()), 0, 0);

    /* Set minimal size */
    g_signal_connect (gobj(), "style-set",
		      G_CALLBACK (tab_label_style_set_cb), this);


    g_signal_connect_after(
      G_OBJECT(document),
      "modified-changed",
      G_CALLBACK(on_modified_changed_static),
      this
    );

    g_signal_connect_after(
      G_OBJECT(document),
      "title-changed",
      G_CALLBACK(on_title_changed_static),
      this
    );

    m_button.add(m_image);
    m_button.set_relief(Gtk::RELIEF_NONE);

    pack_start(m_label, Gtk::PACK_SHRINK);
    pack_end(m_button, Gtk::PACK_SHRINK);
  }

  void TabLabel::on_title_changed()
  {
    update_label();
  }

  void TabLabel::on_modified_changed()
  {
    update_label();
  }

  void TabLabel::update_label()
  {
    bool modified = gtk_text_buffer_get_modified(GTK_TEXT_BUFFER(m_document));
    m_label.set_text((modified ? "*" : "") + std::string(c4s_document_get_title(m_document)));
  }

  bool load_keywords(C4sKeywordList* keywords, GError** error)
  {
    // Use some glib magic to get the folder were the keyword file
    // should be located.
    for(const gchar* const* dirs = g_get_system_data_dirs(); *dirs != NULL; ++ dirs)
    {
      GError* local_error = NULL;
      std::string dir = Glib::build_filename(*dirs, "c4scripter");

      if(c4s_keyword_list_load_from_file(keywords, Glib::build_filename(dir, "keywords.txt").c_str(), &local_error) == FALSE)
      {
        if(local_error->domain == g_file_error_quark() && local_error->code == G_FILE_ERROR_NOENT)
        {
          g_error_free(local_error);
          continue;
        }
        else
        {
          g_propagate_error(error, local_error);
          return false;
        }
      }
      else
      {
        return true;
      }
    }

#ifdef _WIN32
    // Also look at the exe path on Win32
    TCHAR path[MAX_PATH];
    if(!GetModuleFileName(NULL, path, MAX_PATH))
      return false;

    std::string filename = Glib::build_filename(Glib::path_get_dirname(path), "keywords.txt");
    GError* local_error = NULL;

    if(c4s_keyword_list_load_from_file(keywords, filename.c_str(), &local_error) == FALSE)
    {
      if(local_error->domain == g_file_error_quark() && local_error->code == G_FILE_ERROR_NOENT)
      {
        g_error_free(local_error);
      }
      else
      {
        g_propagate_error(error, local_error);
        return false;
      }
    }
    else
    {
      return true;
    }
#endif
    g_set_error(
      error,
      g_file_error_quark(),
      G_FILE_ERROR_NOENT,
      _("Did not find keywords.txt file")
    );

    return false;
  }
}

FileHolder::FileHolder(C4sPreferences* preferences, Gtk::Window& parent):
  Gtk::Notebook(),
  m_preferences(preferences),
  m_parent(parent),
  m_keywords(c4s_keyword_list_new()),
  m_functions(c4s_engine_function_list_new(parent.gobj(), preferences))
{
  GError* error = NULL;
  set_scrollable(true);

  if(load_keywords(m_keywords, &error) == FALSE)
  {
    Gtk::MessageDialog dialog(parent, _("Failed to load keywords"), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    dialog.set_title(_("Could not load keyword file"));
    gchar* text = g_markup_printf_escaped(_("C4Scripter failed to read the keywords.txt file: <b>%s</b>\n\nProbably C4Scripter has not been installed properly. Make sure the keywords.txt is at a location where C4Scripter can find it.\n\nAs a result keyword highlighting will be disabled."), error->message);
    dialog.set_secondary_text(text);
    dialog.property_secondary_use_markup() = true;
    g_free(text);
    g_error_free(error);
    dialog.run();
  }
}

FileHolder::~FileHolder()
{
  g_object_unref(G_OBJECT(m_functions));
  g_object_unref(G_OBJECT(m_keywords));
}

void FileHolder::on_switch_page(GtkNotebookPage* page, guint page_num)
{
  Gtk::Notebook::on_switch_page(page, page_num);

  if(page_num == static_cast<guint>(-1))
  {
    // Is done below in remove_document
    //m_signal_document_changed.emit(NULL);
  }
  else
  {
    C4sDocumentWindow* window = C4S_DOCUMENT_WINDOW(get_nth_page(page_num)->gobj());
    m_signal_document_changed.emit(window);
  }
}

C4sDocumentWindow* FileHolder::create_document()
{
  return create_document(c4s_document_window_new(m_keywords, m_functions, m_preferences));
}

C4sDocumentWindow* FileHolder::create_document(const std::string& path)
{
  C4sDocument* document;
  C4sDocumentWindow* window;
  GError* error;

  error = NULL;
  document = c4s_document_new(m_preferences);
  c4s_document_load_file(document, path.c_str(), &error);
  if(error != NULL) { g_object_unref(G_OBJECT(document)); throw Glib::Error(error, false); }

  window = c4s_document_window_new_with_document(document, m_keywords, m_functions, m_preferences);
  g_object_unref(G_OBJECT(document));
  return create_document(window);
}

C4sDocumentWindow* FileHolder::create_document(C4sDocumentWindow* window)
{
  TabLabel* label = Gtk::manage(new TabLabel(c4s_document_window_get_document(window)));
  label->show_all();

  // TODO: Should just be a close request to allow close confirmation
  label->signal_close().connect(sigc::bind(sigc::mem_fun(*this, &FileHolder::on_document_close), window));

  append_page(*Glib::wrap(GTK_WIDGET(window)), *label);
  gtk_widget_show(GTK_WIDGET(window));

  m_signal_document_added.emit(window);

  set_current_page(get_n_pages() - 1);
  GtkTextIter begin;
  gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(c4s_document_window_get_document(window)), &begin);
  gtk_text_buffer_place_cursor(GTK_TEXT_BUFFER(c4s_document_window_get_document(window)), &begin);
  gtk_widget_grab_focus(GTK_WIDGET(c4s_document_window_get_script_view(window)));

  return window;
}

void FileHolder::remove_document(C4sDocumentWindow* window)
{
  m_signal_document_removed.emit(window);
  remove_page(*Glib::wrap(GTK_WIDGET(window)));

  // Corresponding switch_page signal is not triggered, but we have to inform
  // others when the last page has been removed
  if(get_n_pages() == 0)
    m_signal_document_changed(NULL);
}

C4sDocumentWindow* FileHolder::get_current_document()
{
  Gtk::Widget* widget = get_nth_page(get_current_page() );
  if(widget == NULL) return NULL;
  return C4S_DOCUMENT_WINDOW(widget->gobj());
}

void FileHolder::on_document_close(C4sDocumentWindow* window)
{
  // Called from TabLabel
  m_signal_document_close.emit(window);
}
