/* c4scripter - A GTK+ based C4Script editor
 * Copyright (C) 2006 Bernhard Bonigl, Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <c4scripter/c4s-keywordlist.h>

enum {
  PROP_0,

  PROP_KEYWORDS
};

typedef struct _C4sKeywordListPrivate C4sKeywordListPrivate;
struct _C4sKeywordListPrivate {
  gchar** keywords;
};

#define C4S_KEYWORD_LIST_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE( \
    (o), \
    C4S_TYPE_KEYWORD_LIST, \
    C4sKeywordListPrivate \
  ))

static GObjectClass* parent_class;

static void
c4s_keyword_list_init(C4sKeywordList* keyword_list)
{
  C4sKeywordListPrivate* priv;
  priv = C4S_KEYWORD_LIST_PRIVATE(keyword_list);

  priv->keywords = NULL;
}

static void 
c4s_keyword_list_finalize(GObject* object)
{
  C4sKeywordList* keyword_list;
  C4sKeywordListPrivate* priv;

  keyword_list = C4S_KEYWORD_LIST(object);
  priv = C4S_KEYWORD_LIST_PRIVATE(keyword_list);

  g_strfreev(priv->keywords);

  G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
c4s_keyword_list_set_property(GObject* object,
                              guint prop_id,
                              const GValue* value,
                              GParamSpec* pspec)
{
  C4sKeywordList* keyword_list;
  C4sKeywordListPrivate* priv;

  gchar** keywords;
  guint length;
  guint i;

  g_return_if_fail(C4S_IS_KEYWORD_LIST(object));

  keyword_list = C4S_KEYWORD_LIST(object);
  priv = C4S_KEYWORD_LIST_PRIVATE(keyword_list);

  switch(prop_id)
  {
  case PROP_KEYWORDS:
    keywords = g_value_get_pointer(value);
    length = g_strv_length(keywords);

    if(priv->keywords != NULL)
      g_strfreev(priv->keywords);

    priv->keywords = g_malloc( (length + 1) * sizeof(gchar*));
    for(i = 0; i < length; ++ i)
      priv->keywords[i] = g_strdup(keywords[i]);
    priv->keywords[length] = NULL;

    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    break;
  }
}

static void
c4s_keyword_list_get_property(GObject* object,
                              guint prop_id,
                              GValue* value,
                              GParamSpec* pspec)
{
  C4sKeywordList* keyword_list;
  C4sKeywordListPrivate* priv;

  g_return_if_fail(C4S_IS_KEYWORD_LIST(object));

  keyword_list = C4S_KEYWORD_LIST(object);
  priv = C4S_KEYWORD_LIST_PRIVATE(keyword_list);

  switch(prop_id)
  {
  case PROP_KEYWORDS:
    g_value_set_pointer(value, priv->keywords);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
c4s_keyword_list_class_init(C4sKeywordListClass* klass)
{
  GObjectClass* object_class = G_OBJECT_CLASS(klass);
  parent_class = g_type_class_peek_parent(klass);

  g_type_class_add_private(klass, sizeof(C4sKeywordListPrivate));

  object_class->finalize = c4s_keyword_list_finalize;
  object_class->set_property = c4s_keyword_list_set_property;
  object_class->get_property = c4s_keyword_list_get_property;

  g_object_class_install_property(
    object_class,
    PROP_KEYWORDS,
    /* TODO: Make this a value array */
    g_param_spec_pointer(
      "keywords",
      "C4Script Keywords",
      "NULL-terminated keyword list",
      G_PARAM_READWRITE
    )
  );
}

GType
c4s_keyword_list_get_type(void)
{
  static GType our_type = 0;

  if(our_type == 0)
  {
    static const GTypeInfo our_info = {
      sizeof(C4sKeywordListClass),
      (GBaseInitFunc)NULL,
      (GBaseFinalizeFunc)NULL,
      (GClassInitFunc)c4s_keyword_list_class_init,
      NULL,
      NULL,
      sizeof(C4sKeywordList),
      0,
      (GInstanceInitFunc)c4s_keyword_list_init,
      NULL
    };

    our_type = g_type_register_static(
      G_TYPE_OBJECT,
      "C4sKeywordList",
      &our_info,
      0
    );
  }

  return our_type;
}

C4sKeywordList*
c4s_keyword_list_new(void)
{
  GObject* object;
  object = g_object_new(C4S_TYPE_KEYWORD_LIST, NULL);
  return C4S_KEYWORD_LIST(object);
}

gboolean
c4s_keyword_list_load_from_file(C4sKeywordList* list,
                                const gchar* file,
                                GError** error)
{
  C4sKeywordListPrivate* priv;
  GIOChannel* channel;
  GIOStatus status;
  gchar* line;
  guint keyword_count;

  priv = C4S_KEYWORD_LIST_PRIVATE(list);

  channel = g_io_channel_new_file(file, "r", error);
  if(channel == NULL) return FALSE;

  if(priv->keywords == NULL)
    keyword_count = 0;
  else
    keyword_count = g_strv_length(priv->keywords);

  status = G_IO_STATUS_NORMAL;
  while(status == G_IO_STATUS_NORMAL || status == G_IO_STATUS_AGAIN)
  {
    status = g_io_channel_read_line(
      channel,
      &line,
      NULL,
      NULL,
      error
    );

    if(status == G_IO_STATUS_NORMAL)
    {
      if(line != NULL)
      {
        g_strstrip(line);
        if(line[0] != ';')
        {
          /* TODO: Prealloc */
          ++ keyword_count;
          priv->keywords = g_realloc(
            priv->keywords,
            (keyword_count + 1) * sizeof(gchar*)
          );

          priv->keywords[keyword_count - 1] = line;
          priv->keywords[keyword_count] = NULL;
        }
      }
    }
  }

  g_io_channel_unref(channel);
  if(status == G_IO_STATUS_ERROR) return FALSE;
  return TRUE;
}

gchar* const*
c4s_keyword_list_get_keywords(C4sKeywordList* list)
{
  return C4S_KEYWORD_LIST_PRIVATE(list)->keywords;
}
